// Script for Special Abilities

// FreeSlots for the Luster Purge attack
freeslot("MT_LUSTERPURGE","S_LUSTERPURGE","SPR_LPRG")

local ppsys = CV_RegisterVar({"ppsys", "Off", CV_NETVAR, CV_OnOff})
local mprecharge = CV_RegisterVar({"mprecharge", "4", CV_NETVAR|CV_SHOWMODIF, CV_Natural})

local MAXPP = 10 				//Maximum PP
local PPCHARGERATE = 300 		//PP Charge for 1PP
local THUNDERRADIUS = 320 		//Taken from the k_kart.c file
local ANGLETOFINESHIFT = 19

//Purge states
local SCANNINGPURGE = 1
local HOMINGPURGE = 2
local DETONATINGPURGE = 3

//Move slots and details regarding moves
local EXTREMESPEED = 1 			//Slot for ExtremeSpeed Move
local LUSTERPURGE = 2 			//Slot for Luster Purge Move
local THUNDERWAVE = 3 			//Slot for Thunder Wave Move (Race)
local FLY = 3 					//Slot for Fly Move (Battle)
local DOUBLETEAM = 4		 	//Slot for Double Team Move
local MISTPURGE = 5			 	//Slot for Mist Purge Move
local MAXMOVES = 5 				//Max moves
local LUSTERPURGERANGE = 4098	//Range of the Luster Purge attack before it begins homing in.

local freeplaymode				//Used to check if Freeplay is enabled

local MOVEPPUSAGE = {
	3, //ExtremeSpeed
	4, //Luster Purge
	5, //Thunder Wave/Fly
	6, //Double Team
	MAXPP //Mist Purge, always the same as the Max PP
}

local ITEMPPCHARGE = {
	(PPCHARGERATE/3) * 2, //Sneaker
	(PPCHARGERATE/3) * 5, //Rocket Sneaker
	(PPCHARGERATE/4) * 9, //Invincability
	(PPCHARGERATE/3) * 1, //Banana
	(PPCHARGERATE/3) * 1, //Buttnic Monitor
	(PPCHARGERATE/4) * 1, //Orbinaut
	(PPCHARGERATE/4) * 5, //Jawz
	(PPCHARGERATE/2) * 1, //Mine
	(PPCHARGERATE/3) * 5, //Ballhog
	(PPCHARGERATE/2) * 9, //Blue Bomb
	(PPCHARGERATE/4) * 7, //Grow
	(PPCHARGERATE/2) * 3, //Shrink
	-1, //Thunder Shield
	(PPCHARGERATE/3) * 4, //Hyudoro
	(PPCHARGERATE/3) * 1, //Pogo Spring
	-1 //Kitchen Sink
}

local function PurgeNuke(purge, radius)
	local ns = 60 << FRACBITS
	local mo
	local fa

	for i = 0,15
		fa =  i*ANGLE_22h
		mo = P_SpawnMobj(purge.x, purge.y, purge.z, MT_SUPERSPARK)
		mo.momx = FixedMul(sin(fa),ns)
		mo.momy = FixedMul(cos(fa),ns)
	end
	
	for player in players.iterate do
		if player == purge.user then continue end //Ignore the user of the Purge
		if player and player.valid and player.mo and player.mo.valid then
			if (P_AproxDistance(P_AproxDistance(purge.x - player.mo.x, purge.y - player.mo.y), purge.z - player.mo.z) > radius) then continue end
			if player.mo.health > 0 then
				if not player.kartstuff[k_respawn] then
					K_SpinPlayer(player, purge.user.mo, 0, purge, false)
				end
			end
		end
	end
end

local function FindPurgeTarget(purge)
	//Scans for targets in front of the user within range.
	local targetangle
	local targetchoice
	local targetdistance
	local tempangle
	local dist
	
	targetangle = 45
	
	for player in players.iterate do
		if player == purge.user continue end //Ignore the user of the Purge
		if player and player.valid and player.mo and player.mo.valid then
			if player.mo.health > 0 then
				if not player.kartstuff[k_respawn] then
					//Possible valid target. Get the angle of them relative to the Purge's angle
					tempangle = (purge.angle - R_PointToAngle2(player.mo.x, player.mo.y, purge.x, purge.y)) / ANG1 //This doesn't seem to work as intended?
					if tempangle < 0 then tempangle = $ * -1 end //Get a postive angle
					tempangle = 180 - $ //Somehow, the angle goes the wrong way, so had to invert this
					//print(tempangle)
					if (tempangle < targetangle) then
						//Possible target, but are they in range?
						dist = P_AproxDistance(P_AproxDistance(player.mo.x - purge.x, player.mo.y - purge.y), player.mo.z - purge.z)
						if (dist < LUSTERPURGERANGE * FRACUNIT) then
							//In range, so set them as the target
							targetangle = tempangle
							targetchoice = player
						end
					end
				end
			end
		end
	end	
	
	if targetchoice then return targetchoice else return end
end

function A_LusterPurge(actor, var1, var2)
	//This function does everything for the Luster Purge Attack
	actor.sprite = SPR_LPRG
	local loopframe = actor.user.realtime % 12
	if loopframe < 3 then
		actor.frame = 0
	elseif loopframe < 6 then
		actor.frame = 1
	elseif loopframe < 9 then
		actor.frame = 2
	else
		actor.frame = 3
	end

	if actor.mode == SCANNINGPURGE then

	//Used to scan for a valid target for Luster Purge.
	//If one is found, set that as the target, then go after them.
	//Otherwise, follow the user.
	//The user of the attack is not a valid target.
	
	local dest
	local dast
	dast = FindPurgeTarget(actor)
	if dast then
		//Target found, so fire!
		//print("Target Locked")
		dest = dast.mo
		actor.target = dest
		actor.mode = HOMINGPURGE //Switch to Homing Mode
		actor.user.lpurge = false //No Purge being held, so can use another.
		//print("Mode Switched")
		actor.angle = R_PointToAngle2(actor.x, actor.y, dest.x, dest.y)
		local dist = P_AproxDistance(P_AproxDistance(dest.x - actor.x, dest.y - actor.y), dest.z - actor.z)
		if (dist < (RING_DIST/8)) then 
			//Close enough, so activate detonation
			actor.mode = DETONATINGPURGE
			//print("Detonating - Target in range")
		else
			local speedmul = FixedMul((33 * K_GetKartSpeed(dest.player, false)) / 32, actor.scale)
			actor.momx = FixedMul(FixedDiv(dest.x - actor.x, dist), speedmul)
			actor.momy = FixedMul(FixedDiv(dest.y - actor.y, dist), speedmul)
			actor.momz = FixedMul(FixedDiv(dest.z - actor.z, dist), speedmul)
			//print("Firing")
		end
	else
		//Target not found, so stay with the user.
		if not (actor.user.mo and actor.user.mo.valid) or (actor.user.kartstuff[k_respawn]) then
			//User is not valid or had died, so remove the Purge
			P_RemoveMobj(actor)
			//print("Purge Removed - User died")
		else
			P_InstaThrust(actor, actor.user.mo.angle, 96*FRACUNIT)
			P_TeleportMove(actor, (actor.user.mo.x + actor.momx), (actor.user.mo.y + actor.momy), (actor.user.mo.z + actor.momz))
			actor.angle = actor.user.mo.angle
		end
	end

	elseif actor.mode == HOMINGPURGE then

	//A modified version of A_HomingChase for the Luster Purge.
	//Homes in on the target.
	//If the target dies, the Purge will scan for another valid target.
	//If one is found, go after them. Otherwise, remove the object.
	//The user of the attack is not a valid target.
	local dest
	local dist
	local speedmul
	dest = actor.target
	
	if (not dest) or (not dest.health) then
		//print("Scanning for new target...")
		//Scan for a new target
		local nearestplayer
		local dis
		local tempdis
		dis = FRACUNIT * 8092
		for player in players.iterate do
			if player == actor.user then continue end //Ignore the user
			if player and player.valid and player.mo and player.mo.valid then
				if player.mo.health > 0 then
					if not player.kartstuff[k_respawn] then
						tempdis = P_AproxDistance(actor.x - player.x, actor.y - player.y)
						if tempdis < dis then
							dis = tempdis
							dest = player.mo
						end
					end
				end
			end
		end
		//Check to see if one was found. Otherwise, remove the object
		if (not dest) or (not dest.health) then
			P_RemoveMobj(actor)
			//print("Purge Removed - New Target not found")
			return
		else
			actor.target = dest
			//print("New Target found")
		end
	end
	
	//Got a target, so chase after them
	actor.angle = R_PointToAngle2(actor.x, actor.y, dest.x, dest.y)
	dist = P_AproxDistance(P_AproxDistance(dest.x - actor.x, dest.y - actor.y), dest.z - actor.z)
	if (dist < (RING_DIST/8)) then 
		//Close enough, so activate detonation
		actor.mode = DETONATINGPURGE
		//print("Detonating - Target in range")
	else
		local speedmul = FixedMul((33 * K_GetKartSpeed(dest.player, false)) / 32, actor.scale)
		actor.momx = FixedMul(FixedDiv(dest.x - actor.x, dist), speedmul)
		actor.momy = FixedMul(FixedDiv(dest.y - actor.y, dist), speedmul)
		actor.momz = FixedMul(FixedDiv(dest.z - actor.z, dist), speedmul)
	end
	end
	if actor.mode == DETONATINGPURGE then

	//A modified version of A_RingExplode for the Luster Purge.
	//Planned to not harm the user.
	PurgeNuke(actor, RING_DIST)
	P_RemoveMobj(actor) //Remove the object
	end
end

local function createpurge(player)
	local lpurge = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_LUSTERPURGE)
	P_InstaThrust(lpurge, player.mo.angle, 96*FRACUNIT)
	P_TeleportMove(lpurge, (lpurge.x + lpurge.momx), (lpurge.y + lpurge.momy), (lpurge.z + lpurge.momz))
	lpurge.momx = 0
	lpurge.momy = 0
	lpurge.momz = 0
	lpurge.user = player //Make sure the Purge knows who used it.
	lpurge.mode = SCANNINGPURGE //Set to scanning Mode
	player.lpurge = true //Currently have a Luster Purge prepped, so can't use another
end

local function freeplaycheck()
	local numplayers = 0
	for player in players.iterate do
		if not ((player.spectator) or (not player.mo)) then
			numplayers = $ + 1
			if numplayers > 1 return false end //More than one player, so can stop now
		end
	end
	return true
end

local function thunderwave(player)
	//Temp import from C to Lua
	if player and player.valid and player.mo and player.mo.valid then
	
	local mobj
	local i = 0
	local sx
	local sy
	local an
	
	S_StartSound(player.mo, sfx_zio3)
	P_NukeEnemies(player.mo, player.mo, RING_DIST/4)
	
	mobj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_THOK)
	mobj.state = S_LZIO11
	//mobj.color = SKINCOLOR_TEAL
	mobj.scale = player.mo.scale*3 + (player.mo.scale/2)
	
	mobj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_THOK)
	mobj.state = S_LZIO21
	//mobj.color = SKINCOLOR_CYAN
	mobj.scale = player.mo.scale*3 + (player.mo.scale/2)

	// spawn horizontal bolts;
	
	for i = 1, 7
		mobj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_THOK)
		mobj.angle = P_RandomRange(0, 359)*ANG1
		mobj.fuse = P_RandomRange(20, 50)
		mobj.state = S_KLIT1
	end
	
	an = ANGLE_22h;
	
	for i = 1, 15
		sx = player.mo.x + FixedMul((player.mo.scale*THUNDERRADIUS), cos((an*i)>>ANGLETOFINESHIFT));
		sy = player.mo.y + FixedMul((player.mo.scale*THUNDERRADIUS), sin((an*i)>>ANGLETOFINESHIFT));
		mobj = P_SpawnMobj(sx, sy, player.mo.z, MT_THOK)
		mobj.angle = an*i
		mobj.extravalue1 = THUNDERRADIUS	// Used to know whether we should teleport by radius or something.
		mobj.scale = player.mo.scale*3
		mobj.state = S_KSPARK1
	end
	end
end

local function draw_PP(v, x, y, pp, ppcost, flags)
	local prenum
	local draw_num
	if not pp then pp = 0 end
	
	pp = max($, 0)
	pp = min($, MAXPP)
	
	prenum = ""
	if (pp < 10) and (MAXPP > 9)
		prenum = "0"
	end
	
	local pp_s = prenum..tostring(pp)
	
	prenum = ""
	if (ppcost < 10) and (MAXPP > 9)
		prenum = "0"
	end
	
	local ppcost_s = prenum..tostring(ppcost)
	local num_xpos = x
	local num_ypos = y
	
	if MAXPP > 9
		for i = 1, 2
			draw_num =string.sub(pp_s, i, i)
			if draw_num != " "
				//print(draw_num)
				v.draw(num_xpos, num_ypos, v.cachePatch("STTNUM"..tostring(draw_num)), flags)
			end
			num_xpos = $ + 8
		end
		v.draw(num_xpos, num_ypos, v.cachePatch("STTSLASH"), flags)
		num_xpos = $ + 8
		
		for i = 1, 2
			draw_num =string.sub(ppcost_s, i, i)
			//print(draw_num)
			v.draw(num_xpos, num_ypos, v.cachePatch("STTNUM"..tostring(draw_num)), flags)
			num_xpos = $ + 8
		end
	else
		num_xpos = $ + 8
		draw_num =string.sub(pp_s, 1, 1)
		//print(draw_num)
		v.draw(num_xpos, num_ypos, v.cachePatch("STTNUM"..tostring(draw_num)), flags)
		num_xpos = $ + 8
		
		v.draw(num_xpos, num_ypos, v.cachePatch("STTSLASH"), flags)
		num_xpos = $ + 8
		
		draw_num =string.sub(ppcost_s, 1, 1)
		//print(draw_num)
		v.draw(num_xpos, num_ypos, v.cachePatch("STTNUM"..tostring(draw_num)), flags)
	end
	num_ypos = $ + 3
	v.draw(num_xpos, num_ypos, v.cachePatch("STCFN080"), flags)
	num_xpos = $ + 8
	v.draw(num_xpos, num_ypos, v.cachePatch("STCFN080"), flags)
	
end

addHook("MapLoad", do
	for p in players.iterate do
        if p.pp ~= nil then
            p.pp = 0 //No PP to start
			p.move = 1 //ExtremeSpeed set as the first move
			p.mistpurge = true //Can use Mist Purge
			p.mistpurgerecharge = -1 //Recharge lap set to n/a
			p.mistpurgerestored = 0 //Mist Purge is already active
		end
	end
	
	freeplaymode = freeplaycheck()
end)

addHook("ThinkFrame", function()
	
	if (freeplaymode == nil) or ((leveltime % TICRATE) == 0) then 
		freeplaymode = freeplaycheck()
	end
	
	for player in players.iterate do
		if player.mo.skin ~= "latius" or ppsys.value == 0 then 
			if player.pp then //Remove the PP stuff
				player.pp = nil
				player.move = nil
				player.mistpurge = nil
				player.mistpurgerecharge = nil
				player.mistpurgerestored = nil
			end
			continue //Only if the player is using the Lati skin and the PP system is on
		elseif player.exiting then
			continue //Also, if they're out, they can't use PP and such.
		end 
		
		if player.pp == nil then player.pp = 0 end
		if player.move == nil then player.move = 1 end
		if player.mo.ppcharge == nil then player.mo.ppcharge = 0 end
		if player.mistpurge == nil then
			player.mistpurge = true
			player.mistpurgerecharge = -1
			player.mistpurgerestored = 0
		end
		
		if not (player.cmd.buttons & BT_CUSTOM1) then
			player.moveusetapready = true
			player.moveusetapping = false
		elseif player.moveusetapready then
			player.moveusetapping = true
			player.moveusetapready = false
		else
			player.moveusetapping = false
		end
		
		if player.moveusetapping then //Use Move
			if player.pp > (MOVEPPUSAGE[player.move] - 1)
			and not (player.lpurge and ((player.move == LUSTERPURGE) or (player.move == MISTPURGE)))
			and not (freeplaymode and (player.move == LUSTERPURGE)) then
				//Reduce PP by the move's usage
				player.pp = $ - MOVEPPUSAGE[player.move]
				
				//Slow down PP recharge by 2 seconds
				player.mo.ppcharge = $ - (TICRATE * 2)
				
				//Use the Move
				if player.move == EXTREMESPEED then
					//Boost as if a Sneaker was used
					K_DoSneaker(player, 1) 
					//print("ExtremeSpeed Used")
					
				elseif player.move == LUSTERPURGE then
					//Fires off homing attack
					createpurge(player)
					//print("Luster Purge Used")
				
				elseif (player.move == THUNDERWAVE) and (gametype == 0) then
					//Grants Thunder Shield's activation effects
					thunderwave(player)
					//print("Thunder Wave Used")
					
				elseif player.move == FLY then
					//Grants Pogo Springs effect
					K_DoPogoSpring(player.mo, 0, 1)
					//print("Fly Used")
					
				elseif player.move == DOUBLETEAM then
					//Grants Hyudoros intangability for 10 seconds
					player.kartstuff[k_hyudorotimer] = 10*TICRATE
					//print("Double Team Used")
				
				elseif player.move == MISTPURGE then
					S_StartSound(player.mo, sfx_kgloat)
					//Fires off a SPB and a Luster Purge, as well as granting invincability for 12 seconds
					local new = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, 459)
					P_InstaThrust(new, player.mo.angle, 128*FRACUNIT)
					P_TeleportMove(new, (new.x + new.momx), (new.y + new.momy), (new.z + new.momz + 32 * FRACUNIT))
					new.momx = 0
					new.momy = 0
					new.momz = 0
					
					if not freeplaymode then createpurge(player) end
					
					if not (player.superitemtimer == nil) then //Mist Purge enables Super Form
						player.superitemtimer = $ + TICRATE * (10 + (2* player.kartstuff[k_position]))
						player.supercooldown = 0
					else
						player.kartstuff[k_invincibilitytimer] = TICRATE * (10 + (2* player.kartstuff[k_position]))
						S_ChangeMusic("kinvnc",true,player) //May give unique music for this
					end
					
					
					//print("Mist Purge Used")
					
					//Mist Purge can no longer be used until recharged
					player.move = EXTREMESPEED
					player.mistpurge = false
					player.mistpurgerecharge = player.laps + mprecharge.value
				end
			end
		end
		
		if player.mistpurgerestored > 0 then
			player.mistpurgerestored = $ - 1
		end
		
		if player.mistpurgerecharge == player.laps then
			player.mistpurge = true
			player.mistpurgerestored = TICRATE * 2 //Used to show that Mist Purge is restored
			player.mistpurgerecharge = -1 //Reset to this value so that the recharge doesn't show again
		end
		
		if not (player.cmd.buttons & BT_CUSTOM2) then
			player.moveswitchtapready = true
			player.moveswitchtapping = false
		elseif player.moveswitchtapready then
			player.moveswitchtapping = true
			player.moveswitchtapready = false
		else
			player.moveswitchtapping = false
		end
		
		if player.moveswitchtapping then //Switch Move
			player.move = $ + 1 //Increase the move slot by 1
			if player.move == MAXMOVES and (not player.mistpurge) then
				player.move = EXTREMESPEED //Reset
			elseif player.move > MAXMOVES then
				player.move = EXTREMESPEED //Reset
			end
		end
		
		if not (player.cmd.buttons & BT_CUSTOM3) then
			player.converttapready = true
			player.converttapping = false
		elseif player.converttapready then
			player.converttapping = true
			player.converttapready = false
		else
			player.converttapping = false
		end
		
		if player.converttapping and (player.kartstuff[k_itemtype] > 0) then //Convert Item
			if (ITEMPPCHARGE[player.kartstuff[k_itemtype]] > 0) and (player.pp < MAXPP) then //Only if it can be converted
				player.mo.ppcharge = $ + ITEMPPCHARGE[player.kartstuff[k_itemtype]] //Convert to PP
				player.kartstuff[k_itemamount] = $ - 1
				if player.kartstuff[k_itemamount] == 0 then
					player.kartstuff[k_itemtype] = 0 //Remove Item from the player
				end
			end
		elseif (player.pp < MAXPP) and (not player.spectator) and (player.realtime > 0) then
			if (player.mo.ppcharge < 0) or (K_IsPlayerLosing(player)) or freeplaymode then
				player.mo.ppcharge = $ + 1
			elseif ((player.realtime % TICRATE) == 0) then
				player.mo.ppcharge = $ + 15
			end
		end
		
		while player.mo.ppcharge >= PPCHARGERATE
			player.pp = $ + 1 //Add 1 to the player's PP
			player.mo.ppcharge = $ - PPCHARGERATE //Reduce PP charge by the charge rate
			if player.pp == MAXPP then
				player.mo.ppcharge = 0 //Max Charge, so can't recharge any more
			end
		end
		
		//Compatibility with superform
		if player.superitemtimer then //Only works if superform is included.
			if player.superitemtimer > 0 then
				if player.tapitem and (player.supercooldown == 0)
				and not player.lpurge and not freeplaymode then
					createpurge(player)
					player.superitemtimer = $1 - (TICRATE * 2)
					player.supercooldown = TICRATE
				end
			end
		end
	end
end)

hud.add(function(v, p)
	if p.spectator then return end -- Nope
	if not p.mo then return end -- Double Nope
	if ((p.mo.skin ~= "latius") or (ppsys.value == 0)) then return end -- TRIPPLE NOPE
	
	draw_PP(v, 8, 54, p.pp, MOVEPPUSAGE[p.move], V_SNAPTOTOP|V_SNAPTOLEFT)	//Draw the PP Gauge
	//Draw the selected move
	local movestring = "No Move"
	if p.move == EXTREMESPEED then
		movestring = "ExtremeSpeed"
	elseif p.move == LUSTERPURGE then
		movestring = "Luster Purge"
	elseif (p.move == THUNDERWAVE) and (gametype == 0) then
		movestring = "Thunder Wave"
	elseif p.move == FLY then
		movestring = "Fly"
	elseif p.move == DOUBLETEAM then
		movestring = "Double Team"
	elseif p.move == MISTPURGE then
		movestring = "Mist Purge"
	end
	v.drawString(8, 66, movestring, V_30TRANS|V_SNAPTOTOP|V_SNAPTOLEFT)
	if MOVEPPUSAGE[p.move] <= p.pp
	and not (p.lpurge and ((p.move == LUSTERPURGE) or (p.move == MISTPURGE)))
	and not (freeplaymode and (p.move == LUSTERPURGE)) then
		v.drawString(8, 74, "Cus1: Activate", V_30TRANS|V_SNAPTOTOP|V_SNAPTOLEFT)
	end
	v.drawString(8, 82, "Cus2: Switch", V_30TRANS|V_SNAPTOTOP|V_SNAPTOLEFT)
	if p.kartstuff[k_itemtype] > 0 then
		if ITEMPPCHARGE[p.kartstuff[k_itemtype]] > 0 and (p.pp < MAXPP) then
			v.drawString(8, 90, "Cus3: Convert", V_30TRANS|V_SNAPTOTOP|V_SNAPTOLEFT)
		end
	end
	if p.mistpurgerestored > 0 then
		if ((p.mistpurgerestored % 6) > 2) then
			v.drawString(8, 98, "Mist Purge Recharged", V_10TRANS|V_SNAPTOTOP|V_SNAPTOLEFT)
		end
	end
	
	
end)